
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

/*
 *  dyn.h: base class for nbody systems
 *.............................................................................
 *    version 1:  Dec 1992   Piet Hut, Steve McMillan, Jun Makino
 *    version 2:
 *.............................................................................
 *     This file includes:
 *  1) definition of class dyn
 *.............................................................................
 */

#ifndef  STARLAB_DYN_H
#  define  STARLAB_DYN_H

#include  "util_math.h"
#include  "node.h"
#include  "kepler.h"

#define BIN_INDENT	21	// for print_pert() and print_binary_params()

//#include  "dyn_kepler.h"   NOTE: this file is included at the end instead,
//                           because it needs to know about the dyn class

/*-----------------------------------------------------------------------------
 *  dyn  --  the simplest class of dynamical particles
 *-----------------------------------------------------------------------------
 */
class  dyn : public node
{
    protected:

        // Global variables!

    	static xreal system_time;	// extended-precision time
    	static real real_system_time;	// convenient real copy

    	static bool use_sstar;  // activate single star evolution

	vector  pos;            // position (3-D Cartesian vector)
	vector  vel;            // velocity: (d/dt) pos
	vector  acc;            // acceleration: (d/dt) vel
	kepler * kep;		// actually a pointer to a kepler orbit object

    public:

        dyn(hbpfp the_hbpfp = new_hydrobase, sbpfp the_sbpfp = new_starbase,
	    bool use_stories = true)
	   : node(the_hbpfp, the_sbpfp, use_stories)
	       {
		   pos = vel = acc = 0;
		   kep = NULL;
	       }
 
	virtual ~dyn() {

	    // Note added by Steve 7/7/98:
	    //
	    // Some care is required when deleting the kepler structure,
	    // as binary components in kira share a kepler.  In that case,
	    // when the kepler of the first component is deleted, the
	    // kepler pointer of the other component must also be set
	    // NULL.  Otherwise, an attempt will be made to delete a
	    // nonexistent structure.

	    if (kep) {
		dyn* s = get_binary_sister();

		if (s && s->kep == kep)
		    s->kep = NULL;

		if ((unsigned int) kep != 1)
		    delete kep;		// in case kep is used as a flag

		kep = NULL;
	    }
	}

        void  set_pos(const vector& new_pos)      {pos = new_pos;}
        void  set_vel(const vector& new_vel)      {vel = new_vel;}
        void  set_acc(const vector& new_acc)      {acc = new_acc;}

	void  clear_pos()                         {pos = 0.0;}
	void  clear_vel()                         {vel = 0.0;}
	void  clear_acc()                         {acc = 0.0;}

	inline void  inc_pos(const vector& d_pos) {pos += d_pos;}
	inline void  inc_vel(const vector& d_vel) {vel += d_vel;}
	inline void  inc_acc(const vector& d_acc) {acc += d_acc;}

	inline void  scale_pos(const real scale_factor)  {pos *= scale_factor;}
	inline void  scale_vel(const real scale_factor)  {vel *= scale_factor;}
	inline void  scale_acc(const real scale_factor)  {acc *= scale_factor;}

	inline vector  get_pos()                  {return pos;}
	inline vector  get_vel()                  {return vel;}
	inline vector  get_acc()                  {return acc;}

	inline xreal get_system_time()		{return system_time;}
	inline real get_real_system_time()	{return real_system_time;}
	void set_system_time(xreal t)		{system_time = t;
					         real_system_time = t;}

	inline dyn * get_parent()
	    {return (dyn*) node::get_parent();}
	inline dyn * get_oldest_daughter()
	    {return (dyn*)node::get_oldest_daughter();}
	inline dyn * get_younger_sister()
	    {return (dyn*) node::get_younger_sister();}
	inline dyn * get_elder_sister()
	    {return (dyn*) node::get_elder_sister();}

        inline dyn * get_root()
            {return (dyn*) node::get_root();}
        inline dyn * get_top_level_node()
            {return (dyn*) node::get_top_level_node();}
        inline dyn * get_binary_sister()
            {return (dyn*) node::get_binary_sister();}

	void  calculate_acceleration(dyn *, real);

	inline kepler * get_kepler()		    {return kep;}
	void  set_kepler(kepler * new_kep)	    {kep = new_kep;}

        virtual istream& scan_dyn_story(istream&);
	virtual ostream& print_dyn_story(ostream &s,
					 bool print_xreal = true,
					 int short_output = 0);

        void  to_com();      // transformation to center-of-mass coordinates
        void flatten_node();

	// Declaration of member function defined in dyn_tt.C

	virtual real get_radius();
	virtual void set_radius(real) {};

        bool get_use_sstar()			{return use_sstar;}
	void set_use_sstar(bool u)		{use_sstar = u;}

	// Placeholders (~null in dyn, realized in hdyn)

	virtual bool nn_stats(real energy_cutoff, real kT,
			      vector center, bool verbose,
			      bool long_binary_output = true,
			      int which = 0);
	virtual real print_pert(bool long_binary_output = true,
				int indent = BIN_INDENT);
};

typedef dyn * dynptr;  // to enable dynamic array declarations such as
                       //    dyn** dyn_ptr_array = new dynptr[n];
                       // (note that the following expression is illegal:
                       //    dyn** dyn_ptr_array = new (dyn *)[n];)

typedef vector (dyn::*dyn_VMF_ptr)(void);     // vector member function pointer
typedef void (dyn::*dyn_MF_ptr)(const vector &);     // member function pointer

inline  node * new_dyn(hbpfp the_hbpfp, sbpfp the_sbpfp,
		       bool use_stories = true)
    {return (node *) new dyn(the_hbpfp, the_sbpfp, use_stories);}

inline  dyn * mkdyn(int n, hbpfp the_hbpfp = new_hydrobase,
	                   sbpfp the_sbpfp = new_starbase)
    {return  (dyn *) mk_flat_tree(n, new_dyn, the_hbpfp, the_sbpfp);}

inline  dyn * get_dyn(istream & s, hbpfp the_hbpfp = new_hydrobase,
		                   sbpfp the_sbpfp = new_starbase,
		      		   bool use_stories = true)
    {return  (dyn *) get_node(s, new_dyn, the_hbpfp, the_sbpfp, use_stories);}
    
inline dyn * common_ancestor(dyn * bi, dyn * bj)
    {return (dyn*) common_ancestor((node*)bi, (node*)bj);}

void dbg_message(char*, dyn*);
void dbg_message(char*, dyn*, dyn*);

vector something_relative_to_root(dyn*, dyn_VMF_ptr);

void scale_mass(dyn *, real);
void get_top_level_energies(dyn * b, real eps2, real &potential, real &kinetic);
void scale_virial(dyn *, real, real&, real);
void scale_energy(dyn *, real, real&);

real get_top_level_kinetic_energy(dyn*b);
real get_kinetic_energy(dyn*b);
real get_tidal_energy(dyn* b, real alpha1, real alpha3);

real pot_on_general_node(dyn * bj, dyn * bi, real eps2, bool cm);
void calculate_energies(dyn * root, real eps2,
			real & epot, real & ekin, real & etot,
			bool cm = false);
void print_recalculated_energies(dyn *, int, real, real e_corr = 0);

void compute_com(dyn*, vector&, vector&);
void compute_com(dyn*);

void compute_density(dyn* b,
		     int k = 12,
		     dyn** list = NULL,
		     int n_list = 0);

void merge_low_level_nodes(dyn* b);

// Overloaded functions:

void compute_max_cod(dyn*, vector&, vector&);
void compute_max_cod(dyn*);

void compute_mean_cod(dyn*, vector&, vector&);
void compute_mean_cod(dyn*);

// From lagrad.C:

void compute_mass_radii(dyn*);
void compute_mass_radii_percentiles(dyn*);
void compute_mass_radii_quartiles(dyn*);

typedef bool boolfn(dyn*);
void compute_general_mass_radii(dyn*, int,
				bool nonlin = false,
				boolfn *bf = NULL);

// From sys_stats.C:

bool parse_sys_stats_main(int argc, char *argv[],
			  int  &which_lagr,
			  bool &binaries,
			  bool &short_output,
			  bool &B_flag,
			  bool &calc_e,
			  bool &n_sq,
			  bool &out,
			  bool &verbose);
void check_addstar(dyn* b);
void sys_stats(dyn* root,
	       real energy_cutoff = 1,
	       bool verbose = true,
	       bool binaries = true,
	       bool short_output = false,
	       int which_lagr = 0,
	       bool print_time = false,
	       bool compute_energy = false,
	       bool allow_n_sq_ops = false,
	       void (*compute_energies)(dyn*, real, real&, real&, real&, bool)
			= calculate_energies,
	       void (*dstar_params)(dyn*) = NULL,
	       bool (*print_dstar_stats)(dyn*, bool, vector, bool) = NULL);

// From dyn_stats.C:

real print_binary_params(kepler* k, real m1, real kT,
			 real dist_from_cod,
			 bool verbose = true,
			 bool long_output = true,
			 int init_indent = 0,
			 int indent = BIN_INDENT);

real get_total_energy(dyn* bi, dyn* bj);
real get_period(dyn* bi, dyn* bj);
void get_total_energy_and_period(dyn* bi, dyn* bj, real& E, real& P);

void initialize_kepler_from_dyn_pair(kepler& k, dyn* bi, dyn* bj,
				     bool minimal = false);

void print_binary_from_dyn_pair(dyn* bi, dyn* bj,
				real kT = 0,
				vector center = vector(0,0,0),
				bool verbose = true,
				bool short_output = false);

real print_structure_recursive(dyn* b,
			       void (*dstar_params)(dyn*),
			       int& n_unp, real& e_unp,
			       real kT = 0.0,
			       vector center = vector(0,0,0),
			       bool verbose = true,
			       bool short_output = false,
			       int indent = 0);

real print_structure_recursive(dyn* b,
			       real kT = 0.0,
			       vector center = vector(0,0,0),
			       bool verbose = true,
			       bool short_output = false,
			       int indent = 0);

void compute_core_parameters(dyn*, int, bool, vector&, real&, int&, real&);

// From plot_stars.C:

void plot_stars(dyn * bi,
		int n = 5,
		int k = 3);

#define  put_dyn  put_node

#include  "dyn_kepler.h"

#endif

//=======================================================================//
//  +---------------+        _\|/_        +------------------------------\\
//  |  the end of:  |         /|\         |  inc/dyn.h
//  +---------------+                     +------------------------------//
//========================= STARLAB =====================================\\
 
