
       //=======================================================//    _\|/_
      //  __  _____           ___                    ___       //      /|\
     //  /      |      ^     |   \  |         ^     |   \     //          _\|/_
    //   \__    |     / \    |___/  |        / \    |___/    //            /|\
   //       \   |    /___\   |  \   |       /___\   |   \   // _\|/_
  //     ___/   |   /     \  |   \  |____  /     \  |___/  //   /|\     
 //                                                       //            _\|/_
//=======================================================//              /|\

//  world.h:  Classes and definitions for 4-D trees.
//.............................................................................
//    version 1:  Oct 2000   Steve McMillan
//    version 2:
//.............................................................................
//     This file includes:
//  1) definition of classes segment, worldline, and worldbundle
//.............................................................................

#ifndef  STARLAB_WORLD_H
#  define  STARLAB_WORLD_H

#include "tdyn.h"

real unique_id(int index);
real unique_id(char *name);
real unique_id(tdyn *b);

// Class definitions:

// A worldline is an indexed pointer to the start of a linked list of
// worldline segments.  Each segment consists of a series of events
// (tdyns) along a particle trajectory.  Tree changes result in new
// worldline segments for all particles involved.  The full worldline
// is the entirety of all such segments.
//
// There is presently considerable redundancy in the data stored.
// We will refine the data structures as the package evolves...

class segment {

    private:

	real id;			// global identifier
	tdyn *first_event;		// first event on the list
	tdyn *last_event;		// last event on the list
	real t_start;			// start time of this segment
	real t_end;			// end time of this segment
	segment *next;			// pointer to the next segment

    public:

	segment() {			// empty segment
	    id = -1;
	    first_event = last_event = NULL;
	    t_start = t_end = 0;
	    next = NULL;
	}

	segment(tdyn *b) {		// segment containing a single event
	    id = unique_id(b);
	    first_event = last_event = b;
	    t_start = t_end = b->get_time();
	    next = NULL;
	}

	real get_id()			{return id;}
	tdyn *get_first_event()		{return first_event;}
	tdyn *get_last_event()		{return last_event;}
	real get_t_start()		{return t_start;}
	real get_t_end()		{return t_end;}

	segment *get_next()		{return next;}
	void set_next(segment *s)	{next = s;}

	void add_event(tdyn *b) {
	    if (unique_id(b) == id) {
		last_event->set_next(b);
		b->set_prev(last_event);
		last_event = b;
		t_end = b->get_time();
	    }
	}

	void print(char *label = NULL);
};

class worldline {

    private:

	real id;			// global identifier
	segment *first_segment;		// first segment
	segment *last_segment;		// last segment
	real t_start;			// start time
	real t_end;			// end time

	// Management of tree traversal:

	real t_curr;			// current time
	tdyn *current;			// current node
	tdyn *current_base;		// current base node

	dyn *tree_node;			// pointer to the corresponding node
					// the interpolated tree at time t

    public:

	worldline() {			// empty worldline
	    id = -1;
	    first_segment = last_segment = NULL;
	    t_start = t_end = 0;

	    t_curr = 0;
	    current = current_base = NULL;

	    tree_node = NULL;
	}

	worldline(segment *s) {		// worldline of a single segment
	    id = s->get_id();
	    first_segment = last_segment = s;
	    t_start = s->get_t_start();
	    t_end = s->get_t_end();

	    t_curr = 0;
	    current = current_base = NULL;

	    tree_node = NULL;
	}

	worldline(tdyn *b) {		// worldline of a single event
	    segment *s = new segment(b);
	    id = s->get_id();
	    first_segment = last_segment = s;
	    t_start = s->get_t_start();
	    t_end = s->get_t_end();

	    t_curr = 0;
	    current = current_base = NULL;

	    tree_node = NULL;
	}

	real get_id()			{return id;}
	segment *get_first_segment()	{return first_segment;}
	segment *get_last_segment()	{return last_segment;}
	real get_t_start()		{return t_start;}

	real get_t_end()		{return t_end;}
	void set_t_end(real t)		{t_end = t;}

	real get_t_curr()		{return t_curr;}
	void set_t_curr(real t)		{t_curr = t;}

	tdyn *get_current()		{return current;}
	void set_current(tdyn* b)	{current = b;}

	tdyn *get_current_base()	{return current_base;}
	void set_current_base(tdyn* b)	{current_base = b;}

	dyn *get_tree_node()		{return tree_node;}
	void set_tree_node(dyn* b)	{tree_node = b;}
	void clear_tree_node()		{tree_node = NULL;}

	void add_segment(segment *s) {
	    if (s->get_id() == id) {
		last_segment->set_next(s);
		last_segment = s;
		t_end = s->get_t_end();
	    }
	}
};

typedef worldline *worldlineptr;	// convenient...

// A worldbundle is a group of worldlines, including structures for
// data management purposes.

class worldbundle {

    private:

	worldlineptr *bundle;		// array of worldline pointers
	int	     nw;		// length of the array
	int	     nw_max;		// maximum length of the array
	real	     t_min;		// minimum time
	real	     t_max;		// maximum time

    public:

	worldbundle() {
	    bundle = NULL;
	    nw = nw_max = 0;
	    t_min = VERY_LARGE_NUMBER;
	    t_max = -VERY_LARGE_NUMBER;
	}

	worldbundle(tdyn *b);

	worldlineptr *get_bundle()	{return bundle;}
	worldline *get_worldline(int i)	{return bundle[i];}

	int get_nw()			{return nw;}
	real get_t_min()		{return t_min;}
	real get_t_max()		{return t_max;}

	int find_index(real id);
	int find_index(char *name);
	int find_index(tdyn *b);

	worldline *find_worldline(real id);
	worldline *find_worldline(char *name);
	worldline *find_worldline(tdyn *b);

	void attach(tdyn *b, int verbose = 0);

	void print();
	void print_worldline(char *name, real dt = 0);
};

// Globally visible functions:

worldbundle *read_bundle(istream &s, int verbose = 0);

int count_segments(worldbundle *wb);
int count_events(worldbundle *wb);

tdyn *find_event(tdyn *bn, real t);
void print_event(tdyn *bn, real t);

vector interpolate_pos(tdyn *p, real t, tdyn *bn = NULL);
vector interpolate_vel(tdyn *p, real t, tdyn *bn = NULL);
vector get_pos(tdyn *b, tdyn *bn, real t = -VERY_LARGE_NUMBER);

dyn *create_interpolated_tree(worldbundle *wb, real t, bool vel = false);

#endif
