
//// mkbinary:  Complete the binary formation process by adding binary
////            orbits to an existing binary tree.  It is assumed that
////            the binary tree structure and all masses have already
////            been set (e.g. by mksecondary), and that positions and
////            velocities of all top-level nodes are already known.
////
////            If possible, use the system parameters to scale the
////            binary parameters.
////
//// Options:   -l    lower limit on binary energy [1]
////            -o    specify interpretation of energy limits [1]
////                      1    binary energy
////                      2    binary energy per unit reduced mass
////                      3    binary energy per unit binary mass
////            -s    specify random seed [random from system clock]
////            -u    upper limit on binary energy [1]
////
//// If the total system energy is already known (saved in the snapshot
//// dyn story), then energies are in units of kT.  Otherwise, energies
//// are in absolute units.

//   Steve McMillan, July 1996

#include "dyn.h"

#ifdef TOOLBOX

local void add_dynamics(dyn* cm, real ecc, real energy)
{
    dyn* primary = cm->get_oldest_daughter();
    dyn* secondary = primary->get_younger_sister();

    real m_total = cm->get_mass();
    real m1 = primary->get_mass();
    real m2 = secondary->get_mass();

    // Set internal orbital elements:

    kepler k;

    real peri = 1; // Default value (unimportant unless ecc = 1).
    if (ecc == 1) peri = 0;

    // For now, binary phase is random.

    real mean_anomaly = randinter(-PI, PI);

    make_standard_kepler(k, 0, m_total, energy, ecc,
			 peri, mean_anomaly);
    //    PRC(m_total);
    //    PRC(energy);
    //    PRL(ecc);

    set_random_orientation(k);

    // Set positions and velocities.

    primary->set_pos(-m2 * k.get_rel_pos() / m_total);
    primary->set_vel(-m2 * k.get_rel_vel() / m_total);

    secondary->set_pos(m1 * k.get_rel_pos() / m_total);
    secondary->set_vel(m1 * k.get_rel_vel() / m_total);

}

local void mkbinary(dyn* b, real e_lower, real e_upper, int option)
{
    // Binary energies are drawn from a 1/e distribution,
    // between the specified limits.
    // For now, use thermal eccentricities.

    real efac = e_upper/e_lower;
    for_all_daughters(dyn, b, bi)
	if (bi->get_oldest_daughter()) {

	    dyn* primary = bi->get_oldest_daughter();
	    dyn* secondary = primary->get_younger_sister();
	    real m_total = bi->get_mass();
	    real mu = primary->get_mass() * secondary->get_mass() / m_total;

	    // Choose binary parameters:
	    //
	    // Option 1 ==> limits refer to binary energy.
	    // Option 2 ==> limits refer to binary energy per unit reduced mass.
	    // Option 3 ==> limits refer to binary energy per unit binary mass.
	    //
	    // However, add_dynamics expects energy per unit reduced mass,
	    // so scale the energy in cases 1 and 3.

	    real ecc = sqrt(randinter(0,1));

	    real energy = e_lower * pow(efac, randinter(0,1));
	    
	    if (option == 1)
		energy /= mu;
	    else if (option == 3)
		energy *= m_total / mu;		

//	    PR(m_total);
//	    PR(primary->get_mass());
//	    PRL(secondary->get_mass());
// 	    PRL(energy);
//	    real sma = 0.5 * primary->get_mass() * secondary->get_mass()
//	                   / energy;
//	    PRL(sma);

	    add_dynamics(bi, ecc, energy);
	}
}

local void scale_limits(real& e1, real& e2, int option,
			int N, real M, real E)
{
    real scale = 1.0;

    if (option == 3) {

	// Limits refer to kinetic energy per unit mass.  Scale to -E/M.

	scale = -E / M;

    } else if (option != 2) {

	// Limits refer to energy.  Scale to kT = -(2/3) E/N.

	scale = -E / (1.5*N);
    }

    e1 *= scale;
    e2 *= scale;
}

void main(int argc, char ** argv)
{
    real e_lower = 1.0, e_upper = 10.0;
    int energy_option = 1;
    int random_seed = 0;
    char seedlog[64];

    check_help();

    extern char *poptarg;
    int c;
    char* param_string = "l:s:o:u:";

    while ((c = pgetopt(argc, argv, param_string)) != -1)
	switch(c) {

	    case 'l': e_lower = atof(poptarg);
		      break;
	    case 'o': energy_option= atoi(poptarg);
		      break;
	    case 's': random_seed = atoi(poptarg);
		      break;
	    case 'u': e_upper = atof(poptarg);
		      break;
            case '?': params_to_usage(cerr, argv[0], param_string);
	    	      get_help();
		      exit(1);
	}

    if (e_lower <= 0 || e_upper <= 0 || e_lower > e_upper)
	err_exit("mkbinary: Illegal limits on semi-major axis");

    dyn* b;
    b = get_dyn(cin);

    b->log_history(argc, argv);

    int actual_seed = srandinter(random_seed);

    sprintf(seedlog, "         random number generator seed = %d",actual_seed);
    b->log_comment(seedlog);

    // Look in the dyn story to see if a total system energy is known.
    // If it is, scale e_lower and e_upper accordingly.

    // Thus, if the energy is known, e_lower and e_upper are specified
    // in "kT" units.  If the energy is not known, e_lower and e_upper
    // are absolute limits on the binary kinetic energy.

    char* energy_string = "total_energy";

    if (find_qmatch(b->get_dyn_story(), energy_string))
	scale_limits(e_lower, e_upper, energy_option,
		     b->n_daughters(), b->get_mass(),
		     getrq(b->get_dyn_story(), energy_string));
    else
	cerr << "mkbinary: Using unscaled energy limits.\n";

    mkbinary(b, e_lower, e_upper, energy_option);

    put_dyn(cout, *b);
    rmtree(b);
}
#endif
