//
// horizontal_branch.C
//

#include "horizontal_branch.h"
#include "sub_giant.h"
#include "hertzsprung_gap.h"

// ANSI C++ first creates the base class before the dreived classes are
// created.
//

horizontal_branch::horizontal_branch(sub_giant & g) : single_star(g) {

    delete &g;

// (GN+SPZ May  4 1999) last update age is time of previous type change
  last_update_age = next_update_age;

    adjust_next_update_age();

    instantaneous_element();
    update();

    post_constructor();
}

horizontal_branch::horizontal_branch(hertzsprung_gap & h) : single_star(h) {

    delete &h;

// (GN+SPZ May  4 1999) last update age is time of previous type change
  last_update_age = next_update_age;

    adjust_next_update_age();

    instantaneous_element();
    update();

    post_constructor();
      
}

#if 0
void horizontal_branch::adjust_initial_star() {

  if(relative_age<=0) {
    real t_ms = main_sequence_time();
    real t_giant = t_ms + hertzsprung_gap_time(t_ms)
                 + base_giant_branch_time(t_ms);
    relative_age = max(t_giant, relative_age);
  }
}
#endif


void horizontal_branch::instantaneous_element() {

    luminosity = min(helium_giant_luminosity(), 
		     maximum_luminosity());

    radius = (0.25*pow(luminosity, 0.4)
           + 0.8*pow(luminosity, 0.67))/pow(relative_mass, 0.27);

    real t_ms = main_sequence_time();
    real t_giant = t_ms + hertzsprung_gap_time(t_ms)
                 + base_giant_branch_time(t_ms);
    real t_he = helium_giant_time(t_ms);
  
    real ff = (relative_age - t_giant)/t_he;
    if (radius<=25)
      radius *= 1.0 - 0.1*ff;
    else
      radius *= pow(25.0/radius, ff); // Giant

    effective_radius = radius;
}

// evolve a horizontal_branch star upto time argument according to
// the model discribed by Eggleton et al. 1989.
void horizontal_branch::evolve_element(const real end_time) {

      real dt = end_time - current_time;
      current_time = end_time;
      relative_age += dt;

      if (relative_age<=next_update_age) {
         luminosity = min(helium_giant_luminosity(), 
                          maximum_luminosity());

	 evolve_core_mass(dt);
	 
         radius = (0.25*pow(luminosity, 0.4)
                + 0.8*pow(luminosity, 0.67))/pow(relative_mass, 0.27);

         real t_ms = main_sequence_time();
         real t_giant = t_ms + hertzsprung_gap_time(t_ms)
                             + base_giant_branch_time(t_ms);
         real t_he = helium_giant_time(t_ms);

         real ff = (relative_age - t_giant)/t_he;
         if (radius<=25)
            radius *= 1.0 - 0.1*ff;
         else
            radius *= pow(25.0/radius, ff); // Giant
      }
      else {
// (GN+SPZ May  4 1999) fudge to fit CO core of super giant
// see helium_star::helium_core_mass which is the CO core of helium star
// for CO core in theory (m_core), but this can not be smaller than 0.55
// massive stars still have hlium core
	if (relative_mass < cnsts.parameters(super_giant2neutron_star)) {
	  real fit_mass = min(2.2, max(core_mass, 0.5));
	  real m_final = 1 - 0.18 * (fit_mass - 0.5);
	  
	  real m_core = m_final * core_mass
	              * cnsts.parameters(helium_star_lifetime_fraction);
	  
	  core_mass = max(0.5, min(core_mass, m_core));
	}

	star_transformation_story(Super_Giant);
	new super_giant(*this);
         return;
      }

      update();
      stellar_wind(dt);
   }

star* horizontal_branch::reduce_mass(const real mdot) {

      if (envelope_mass<=mdot) {
         envelope_mass = 0;
         star_transformation_story(Helium_Star);
         return dynamic_cast(star*, new helium_star(*this));
      }

      envelope_mass -= mdot;
      return this;
   }

star* horizontal_branch::subtrac_mass_from_donor(const real dt,
						 real& mdot) {

      real mdot_temp = relative_mass*dt/get_binary()->get_donor_timescale();
      mdot = mass_ratio_mdot_limit(mdot_temp);

      if (envelope_mass<=mdot) {
         mdot = envelope_mass;
         envelope_mass = 0;
         star_transformation_story(Helium_Star);
         return dynamic_cast(star*, new helium_star(*this));
      }

      envelope_mass -= mdot;
      return this;
   }

void horizontal_branch::adjust_accretor_age(const real mdot, const bool rejuvenate=true) {

      real m_rel_new;
      real m_tot_new = get_total_mass() + mdot;
      if (m_tot_new>relative_mass)
         m_rel_new = m_tot_new;
      else m_rel_new = relative_mass;

      real t_ms = main_sequence_time();
      real t_giant_old = t_ms + hertzsprung_gap_time(t_ms) 
                              + base_giant_branch_time(t_ms);
      real t_he_old = helium_giant_time(t_ms);

      t_ms = main_sequence_time(m_rel_new);
      real t_giant_new = t_ms + hertzsprung_gap_time(m_rel_new, t_ms) 
                              + base_giant_branch_time(m_rel_new, t_ms);
      real t_he_new = helium_giant_time(m_rel_new, t_ms); 

      real dtime = relative_age - t_giant_old;

// (GN+SPZ May  4 1999) update last_update_age
      last_update_age = t_giant_new;
      relative_age = t_giant_new
                   + dtime*(t_he_new/t_he_old);
      if (rejuvenate)
         relative_age *= rejuvenation_fraction(mdot/m_tot_new);

      relative_age = max(relative_age, 
			 last_update_age + cnsts.safety(minimum_timestep));

      // next_update_age should not be reset here
      // next_update_age = t_giant_new + t_he_new;

   }

real horizontal_branch::zeta_adiabatic() {
      return 15;	
   }

real horizontal_branch::zeta_thermal() {
      return 15;
   }

void horizontal_branch::adjust_next_update_age() {
     
      real t_ms = main_sequence_time();
      real t_giant = t_ms + hertzsprung_gap_time(t_ms)
                          + base_giant_branch_time(t_ms);
      real t_he = helium_giant_time(t_ms);
      next_update_age = t_giant + t_he;
   }

#if 0
// Determine size of stellar core according to
// Iben, I, Jr., \& Tutukov, A.V., 1993, ApJ 418, 343. Eq. 5.
// and
// Beveren, van D., 1980, PhD thesis, {\em Free University Brussels}.
real horizontal_branch::helium_core_mass() {

      real m_core, m1_core;
//              Iben and Tutukov 1993
      m_core = (log10(luminosity) - 1.55)/3.56;

//               Beveren 1980
      m1_core = 0.073*(1 + cnsts.parameters(core_overshoot))
	      * pow(relative_mass, 1.42);
      m_core = max(m_core, m1_core);

//              Dewey, R.J., Cordes, J.M., 1987, ApJ 321, 780.
//              Used a minimum helium core mass of \sim 3 M_\odot.
//              But this is not realistic.
//      	if(m_core>=2.8 && m_core<3.1) m_core = 3.1;
//		But we do not beleve them.

//		Limit core_mass by realistic limits.
      //      if(m_core>=M_HE_NS && m_core<M_HEmin) m_core = M_HEmin;
      m_core = min(m_core, get_total_mass());

      return max(m_core, core_mass);
   }

real horizontal_branch::stellar_radius(const real mass, const real age) {

      real t_ms = main_sequence_time(mass);
      real t_giant = t_ms + hertzsprung_gap_time(mass, t_ms)
                          + base_giant_branch_time(mass, t_ms);
      real t_he = helium_giant_time(mass, t_ms);

      real l_hb = min(helium_giant_luminosity(mass),
                      maximum_luminosity(mass));

      real r_hb = (0.25*pow(l_hb, 0.4)
                + 0.8*pow(l_hb, 0.67))/pow(mass, 0.27);

//              Radius adjustment for giants.
//              Some stars evolve from the giant branch to the
//              horizontal_branch and back to the giant granch.
//              Do not transform the star back to the giant branch but
//              adjusts its radius.
//              This may result in a Roche-lobe filling `horizontal_branch'.
      real ff = (age - t_giant)/t_he;
      if (r_hb<=25)
         r_hb *= 1.0 - 0.1*ff;
      else
         r_hb *= pow(25.0/r_hb, ff); // Giant

      return r_hb;
   }
#endif


real horizontal_branch::gyration_radius_sq() {

  return cnsts.parameters(radiative_star_gyration_radius_sq); 
}

void horizontal_branch::evolve_core_mass(const real dt) {

  real delta_mcore = luminosity*dt
                   / cnsts.parameters(energy_to_mass_in_internal_units);
  delta_mcore = min(delta_mcore, 
		    envelope_mass - cnsts.safety(minimum_mass_step));

  if (delta_mcore<0) {
      PRC(luminosity);PRC(dt);PRL(delta_mcore);
      cerr << "Core mass increas is negative in horizontal_branch::evolve_core_mass()"<<endl;
      delta_mcore = 0;
      return;
  }

  envelope_mass -= delta_mcore; 
  core_mass += delta_mcore;
      
}

// wind_constant is fraction of envelope lost in nuclear lifetime
// of stars. Should be updated after mass accretion
// (SPZ+GN: 1 Oct 1998)
void horizontal_branch::update_wind_constant() {
  
// (GN+SPZ Apr 28 1999) new fits to Maeder, de Koter and common sense

  if (relative_mass >= cnsts.parameters(super_giant2neutron_star)) {

    real meader_fit_dm = 0.01*pow(relative_mass,2.);
    
    if (relative_mass < 85)
      wind_constant = meader_fit_dm;
    else {// constant
      real final_mass = 30;
      wind_constant = relative_mass - final_mass;
    }

  } 
  else { // 5% of initial envelope

    wind_constant = 0.05*(relative_mass - final_core_mass());
  }
  
  wind_constant = max(wind_constant, 0.0);

}
